#include "client.hpp"
#include "definitions.hpp"
#include "util.hpp"
#include "wordle-de.pb.h"
#include <cstdio>
#include <iostream>
#include <mutex>
#include <stdexcept>
#include <string>
#include <thread>

using namespace wordle_de;

std::string getUsernameFromInput() {
  std::string username;
  std::getline(std::cin, username);
  if (username.empty()) {
    throw "empty username is not allowed";
  }
  if (username.length() > USERNAME_MAX_LENGTH) {
    throw "username is too long";
  }
  return username;
}

std::string getPasswordFromInput() {
  std::string password;
  std::getline(std::cin, password);
  if (password.empty()) {
    throw "empty password is not allowed";
  }
  if (password.length() < 8) {
    throw "password is too short";
  }
  return password;
}

std::string getGuessFromInput() {
  std::string guess;
  std::getline(std::cin, guess);
  if (guess.empty()) {
    throw "empty guess is not allowed";
  }
  if (guess.length() != WORD_LENGTH) {
    try {
      auto errMsg = string_format("word has to be exactly %d characters long",
                                  WORD_LENGTH);
      throw errMsg;
    } catch (std::runtime_error e) {
      throw "word has not the right length";
    }
  }
  return guess;
}

void received(uint16_t fd, char *inputBuffer) {
  auto msg = wordle_de::Message();
  msg.ParseFromString(inputBuffer);

  switch (msg.msg_case()) {
  case Message::kGuess:
    std::cout << "Guess" << std::endl;
    break;
  case wordle_de::Message::kGameState: {
    std::cout << "Received Game State" << std::endl;
    auto gs = msg.gamestate();
    if (gs.newword()) {
      std::cout << "There is a new word available, resetting game..."
                << std::endl;
    }
    break;
  }
  default:
    break;
  }
}

enum State {
  INPUT_LOGIN,
  WF_LOGIN_ACK,
  INPUT_GUESS,
  WF_GUESS_ACK,
  WF_NEW_WORD
};
int main() {
  GOOGLE_PROTOBUF_VERIFY_VERSION;
  // try to connect to server
  auto cl = Client();
  std::cout << "Connected to server!" << std::endl;
  std::cout << "Logging in..." << std::endl;

  // State machine
  State state = INPUT_LOGIN;
  std::string username;
  while (true) {
    switch (state) {
    case INPUT_LOGIN: {
      std::cout << "Willkommen zu Wordle auf Deutsch!" << std::endl;
      std::cout << "Bitte gib deinen Benutzernamen und Passwort ein:"
                << std::endl;
      std::string password;
      while (true) {
        try {
          std::cout << "Benutzername: ";
          username = getUsernameFromInput();
          std::cout << "Passwort: ";
          password = getPasswordFromInput();
          break;
        } catch (const char *error) {
          std::cout << "Error: " << error << std::endl;
        }
      }
      auto msg = Message();
      auto login = new Login();
      login->set_username(username);
      login->set_password(password);
      msg.set_allocated_login(login);
      cl.sendMessage(msg.SerializeAsString().c_str());
      state = WF_LOGIN_ACK;
      break;
    }
    case WF_LOGIN_ACK: {
      auto msg = cl.receiveMessage();
      if (msg.msg_case() == wordle_de::Message::kLoginAck) {
        if (msg.loginack().loggedin()) {
          std::cout << "Logged in as " << username << std::endl;
          state = INPUT_GUESS;
          break;
        } else {
          std::cout << "Login failed. Error: " << msg.loginack().error()
                    << std::endl;
          state = INPUT_LOGIN;
          break;
        }
      } else {
        std::cout << "Received message unsuspected type." << std::endl;
      }
      state = INPUT_GUESS;
      break;
    }
    case INPUT_GUESS: {
      std::string guess;
      while (true) {
        try {
          std::cout << "Bitte gib ein Wort ein" << std::endl;
          guess = getGuessFromInput();
          break;
        } catch (std::string error) {
          std::cout << "Error: " << error << std::endl;
        }
      }
      auto guessMsg = new wordle_de::Guess();
      guessMsg->set_guess(guess);
      auto msg = wordle_de::Message();
      msg.set_allocated_guess(guessMsg);
      cl.sendMessage(msg.SerializeAsString().c_str());
      state = WF_GUESS_ACK;
    }
    case WF_GUESS_ACK: {
      auto msg = cl.receiveMessage();
      if (msg.msg_case() == wordle_de::Message::kGuessAck) {
        auto result = msg.guessack().result();
        bool correct = true;
        std::string resultOut;
        for (const auto &c : result) {
          switch (c) {
          case wordle_de::GuessAck_matchType::GuessAck_matchType_NO_MATCH:
            correct = false;
            resultOut += "x";
            break;
          case wordle_de::GuessAck_matchType::GuessAck_matchType_MATCH:
            correct = false;
            resultOut += "m";
            break;
          case wordle_de::GuessAck_matchType::GuessAck_matchType_EXACT_MATCH:
            resultOut += "M";
            break;
          }
        }
        std::cout << resultOut << std::endl;
        if (correct) {
          std::cout
              << "Du hast das Rätsel gelöst! Warte auf das nächste Wort..."
              << std::endl;
          state = WF_NEW_WORD;
        } else {
          std::cout << "Versuchs noch einmal!" << std::endl;
          state = INPUT_GUESS;
        }
      } else {
        std::cout << "Received message unsuspected type." << std::endl;
      }
    }
    default:
      break;
    }
  }
}
