#ifndef _SERVER_H
#define _SERVER_H

#include "game.hpp"
#include "wordle-de.pb.h"
#include <functional>
#include <netinet/in.h>
#include <unistd.h>
#include <vector>

#define INPUT_BUFFER_SIZE 1000

class Server {
public:
  typedef std::function<void(uint16_t, char *)> receiveCallbackType;
  typedef std::function<void(uint16_t)> newConnectionCallbackType;
  typedef std::function<void(uint16_t)> disconnectCallbackType;

private:
  Server();
  Server(const int &);
  Server(const Server &) = delete;
  virtual ~Server();

public:
  static void createInstance();
  static void createInstance(const int &);
  static void destroyInstance() { delete instance; }
  static Server *getInstance() { return instance; }
  void init();
  void run();
  void stop();

  // callback setters
  void onConnect(newConnectionCallbackType cb);
  void onInput(receiveCallbackType cb);
  void onDisconnect(disconnectCallbackType cb);

  uint16_t sendMessage(int source_fd, const char *messageBuffer);
  uint16_t sendMessage(int source_fd, char *messageBuffer);

  void sendLoginAck(uint16_t fd, bool success, const char *errorMsg = "") {
    auto ack = new wordle_de::LoginAck();
    ack->set_loggedin(success);
    ack->set_error(errorMsg);
    auto msg = wordle_de::Message();
    msg.set_allocated_loginack(ack);
    sendMessage(fd, msg.SerializeAsString().c_str());
  }
  void sendGuessAck(uint16_t fd, const GameState::guessResultT &res) {
    auto result = new wordle_de::GuessAck();
    for (const auto &r : res) {
      switch (r) {
      case GameState::NO_MATCH:
        result->add_result(
            wordle_de::GuessAck_matchType::GuessAck_matchType_NO_MATCH);
        break;
      case GameState::MATCH:
        result->add_result(
            wordle_de::GuessAck_matchType::GuessAck_matchType_MATCH);
        break;
      case GameState::EXACT_MATCH:
        result->add_result(
            wordle_de::GuessAck_matchType::GuessAck_matchType_EXACT_MATCH);
        break;
      }
    }
    auto msg = wordle_de::Message();
    msg.set_allocated_guessack(result);
    sendMessage(fd, msg.SerializeAsString().c_str());
  }

private:
  // file descriptor sets
  fd_set masterfds;
  fd_set tmpfds;

  // track the maximum fd value
  uint16_t maxfd;

  int mastersocket_fd;
  int tempsocket_fd;

  struct sockaddr_storage client_addr;
  struct sockaddr_in server_addr;

  char input_buffer[INPUT_BUFFER_SIZE];

  char remote_ip[INET6_ADDRSTRLEN];

  newConnectionCallbackType newConnectionCallback;
  receiveCallbackType receiveCallback;
  disconnectCallbackType disconnectCallback;

  // function prototypes
  void setup(int port);
  void initializeSocket();
  void bindSocket();
  void startListen();
  void handleNewConnection();
  void recvInputFromExisting(int fd);

private:
  static Server *instance;
};
#endif // !_SERVER_H
